// cBuilding.cpp

#include "cBuilding.h"
#include <GL/gl.h>
#include "macros.h"


cBuilding::cBuilding(int x, int y, int z, int rooms_x, int rooms_y, int rooms_z) {
  mRoles |= (1UL << cObject::COLLIDEABLE);
  mRoles |= (1UL << cObject::DAMAGEABLE);
  mRoles |= (1UL << cObject::BUILDING);
  mPos[0] = x * 3;
  mPos[1] = y * 3.5;
  mPos[2] = z * 3;
  mRot[0] = mRot[1] = mRot[2] = 0;
  mRooms[0] = 3*rooms_x;
  mRooms[1] = 3.5*rooms_y;
  mRooms[2] = 3*rooms_z;
  this->mRadius = sqrt(0.25*mRooms[0]*mRooms[0] + 0.25*mRooms[1]*mRooms[1] + 0.25*mRooms[2]*mRooms[2]);
}

cObject::rImpact* cBuilding::damageByCylinder(float* a, float* b, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = this->mPos;
    float* r = this->mRot;
    float* w = this->mRooms;
    glPushMatrix();
        glLoadIdentity();
        glRotatef(-r[1], 0,1,0);
        glTranslatef(-p[0],-p[1],-p[2]);
        float M[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, M);
        float a_local[3];
        applyMatrix3(a_local, M, a);
        float b_local[3];
        applyMatrix3(b_local, M, b);
        float mins[3] = { 0, 0, 0 };
        float maxs[3] = { +w[0], +w[1], +w[2] };
        float impact[3];
        float normal[3];
        float delta = collideLineWithAAB(a_local, b_local, mins, maxs, impact, normal);
        if (delta >= 0.0f) {
            printf("Building hit\n");
            this->mHp -= damage;
            if (this->mHp <= 0) {
                mExplosion.fire(NULL);
            }
            result = new cObject::rImpact;
            memcpy(result->pos, impact, sizeof(float)*3);
            memset(result->nrm, 0, sizeof(float)*3);
        }
    //printf("delta %.6f\n", delta);
    glPopMatrix();
//    printVector(a);
//    printVector(a_local);
//    printVector(b);
//    printVector(b_local);
    return result;
}


void cBuilding::onSpawn() {
    mExplosion.mOwner = this;
}


void cBuilding::animateSelf(float spf) {
    mExplosion.animate(spf);
    if (mExplosion.ready() && mHp <= 0) {
        //cWorld::instance->fragObject(this);
    }
}


void cBuilding::transformSelf() {
    mExplosion.transform();
}


void cBuilding::drawSolid() {
    mExplosion.drawSolid();
    
    float* p = this->mPos;
    float* r = this->mRot;
    float* w = this->mRooms;
    glPushAttrib(GL_CURRENT_BIT);
    glPushMatrix();
        glTranslatef(p[0],p[1],p[2]);
        glRotatef(r[1], 0,1,0);

        glTranslatef(0.5*w[0], 0, 0.5*w[2]);
        
        glPushMatrix();
            glScalef(0.5*w[0]-1, 0.5*w[1], 0.5*w[2]-1);
            glTranslatef(0, 1, 0);
            glColor4f(0.2,0.2,0.2,1);
            glCenterUnitBlock();
        glPopMatrix();
        
        int stories = w[1]/3.5;
        for (int i = 0; i < stories; i++) {
            const float low = 0.4;
            const float mid = 2.5;
            glPushMatrix();
                glScalef(0.5*w[0]+0.4, low, 0.5*w[2]+0.4);
                glTranslatef(0, 1, 0);
                glColor4f(0.5,0.5,0.5,1);
                glCenterUnitBlock();
            glPopMatrix();
            glPushMatrix();
                glTranslatef(0,low+mid,0);
                glScalef(0.5*w[0]+0.2, 0.4, 0.5*w[2]+0.2);
                glTranslatef(0, 1, 0);
                glColor4f(0.3,0.3,0.3,1);
                glCenterUnitBlock();
            glPopMatrix();
            glTranslatef(0, 3.5, 0);
        }
        
    glPopMatrix();
    glPopAttrib();
};

void cBuilding::drawEffect() {
    mExplosion.drawEffect();
    
    if (mHp <= 0) return;
    
    float* p = this->mPos;
    float* r = this->mRot;
    float* w = this->mRooms;
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT);
    glDisable(GL_CULL_FACE);
    glEnable(GL_DITHER);
    
    glPushMatrix();
        glRotatef(r[1], 0,1,0);
        glTranslatef(p[0],p[1],p[2]);
        glTranslatef(0.5*w[0], 0, 0.5*w[2]);
        
        glPushMatrix();
            glScalef(0.5*w[0], 0.5*w[1], 0.5*w[2]);
            glTranslatef(0, 1, 0);
            glColor4f(0.4,0.6,0.9,0.4f);
            glCenterUnitBlock();
        glPopMatrix();

        glPushMatrix();
            glScalef(0.5*w[0]+0.1, 0.5*w[1], 0.5*w[2]+0.1);
            glTranslatef(0, 1, 0);
            glColor4f(0.5,0.7,0.9,0.4f);
            glCenterUnitBlock();
        glPopMatrix();
        
    glPopMatrix();
    glPopAttrib();
}




cTile::cTile(int x, int y, int z, int kind) {
    mRoles |= (1UL << cObject::COLLIDEABLE);
    mRoles |= (1UL << cObject::DAMAGEABLE);
    mPos[0] = x * 9.0f;
    mPos[1] = y * 3.0f;
    mPos[2] = z * 9.0f;
    mRot[0] = mRot[1] = mRot[2] = 0.0f;
    mKind = kind;
}


cObject::rImpact* cTile::damageByCylinder(float* a, float* b, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = this->mPos;
    float* r = this->mRot;
    glPushMatrix();
        glLoadIdentity();
        glRotatef(-r[1], 0,1,0);
        glTranslatef(-p[0],-p[1],-p[2]);
        float M[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, M);
        float a_local[3];
        applyMatrix3(a_local, M, a);
        float b_local[3];
        applyMatrix3(b_local, M, b);
        float mins[3] = { 0, -0.1, 0 };
        float maxs[3] = { 9, 0, 9 };
        float impact[3];
        float normal[3];
        float delta = collideLineWithAAB(a_local, b_local, mins, maxs, impact, normal);
        if (delta >= 0.0f) {
            printf("Tile hit\n");
            this->mHp -= damage;
            if (this->mHp <= 0) {
                printf("Tile damaged beyond repair");
            }
            result = new cObject::rImpact;
            memcpy(result->pos, impact, sizeof(float)*3);
            memset(result->nrm, 0, sizeof(float)*3);
        }
    //printf("delta %.6f\n", delta);
    glPopMatrix();
//    printVector(a);
//    printVector(a_local);
//    printVector(b);
//    printVector(b_local);
    return result;
}


void cTile::drawSolid() {
    float* p = this->mPos;
    float* r = this->mRot;
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT);
    glDisable(GL_CULL_FACE);
    glEnable(GL_DITHER);
    
    glPushMatrix();
        glRotatef(r[1], 0,1,0);
        glTranslatef(p[0],p[1],p[2]);
        
        if (mKind == KIND_CONCRETE) {
            glPushMatrix();
                glScalef(9, 0.05, 9);
                glColor4f(0.7,0.7,0.7,0.4f);
                glUnitBlock();
            glPopMatrix();
            
        } else if (mKind == KIND_ROAD_JUNKTION) {

            glPushMatrix();
                glScalef(2, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            
            glPushMatrix();
                glTranslatef(7,0,0);
                glScalef(2, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();

            glPushMatrix();
                glTranslatef(7,0,7);
                glScalef(2, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();

            glPushMatrix();
                glTranslatef(0,0,7);
                glScalef(2, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            
            glPushMatrix();
                glScalef(9, 0.02, 9);
                glColor4f(0.2,0.2,0.2,1);
                glUnitBlock();
            glPopMatrix();
            
        } else if (mKind == KIND_ROAD_NORTHSOUTH) {
            glPushMatrix();
                glScalef(2, 0.05, 9);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            glPushMatrix();
                glTranslatef(2,0,0);
                glScalef(5, 0.02, 9);
                glColor4f(0.2,0.2,0.2,1);
                glUnitBlock();
            glPopMatrix();
            glPushMatrix();
                glTranslatef(7,0,0);
                glScalef(2, 0.05, 9);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            
        } else if (mKind == KIND_ROAD_EASTWEST) {
            glPushMatrix();
                glScalef(9, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            glPushMatrix();
                glTranslatef(0,0,2);
                glScalef(9, 0.02, 5);
                glColor4f(0.2,0.2,0.2,1);
                glUnitBlock();
            glPopMatrix();
            glPushMatrix();
                glTranslatef(0,0,7);
                glScalef(9, 0.05, 2);
                glColor4f(0.7,0.7,0.7,1);
                glUnitBlock();
            glPopMatrix();
            
        }
        
    glPopMatrix();
    glPopAttrib();
}
