// cController.cpp
//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//

#include "cController.h"
#include "cObject.h"

#include <cstdlib>
#include <cassert>

using namespace std;

#define debug !true

//--------------------------------


cController::cController(cObject* entity, bool enabled)
{
  mDevice = entity;
  mEnabled = enabled;
  clearInstruction();
}


void cController::clearInstruction()
{
  if (debug) printf("clearInstruction()\n");
  //  overideInstruction(WAIT, (char*) NULL);
}


void cController::overideInstruction(int opcode, cObject *entity)
{
  if (debug) printf("overideInstruction(%s)\n", getInstructionName(opcode));

  if (entity == NULL) return;
  if (entity->mFragged) return;

  rInstruction* st;

  if (!mInstructions.empty()) {
    st = mInstructions.top();
    if (st->entity_name != NULL) delete st->entity_name;
    mInstructions.pop();
    delete st;
  };

  st = new rInstruction;
  st->opcode = opcode;
  st->entity = entity;
  if (entity->mName != NULL) st->entity_name = new string(*(entity->mName));
  mInstructions.push(st);
}


void cController::overideInstruction(int opcode, char* entityname)
{
  if (debug) printf("overideInstruction(%s, %s)\n", getInstructionName(opcode), entityname);

  if (entityname == NULL) return;

  rInstruction* st;

  if (!mInstructions.empty())
  {
    st = mInstructions.top();
    if (st->entity_name != NULL) delete st->entity_name;
    mInstructions.pop();
    delete st;
  };

  st = new rInstruction;
  st->opcode = opcode;
  st->entity = NULL;
  st->entity_name = new string(entityname);
  mInstructions.push(st);
}


void cController::newInstruction(int opcode, cObject *entity)
{
  if (debug) printf("newInstruction(%s)\n", getInstructionName(opcode));

  if (entity == NULL) return;
  else if (entity->mFragged) return;

  rInstruction* st;

  st = new rInstruction;
  st->opcode = opcode;
  st->entity = entity;
  if (entity->mName != NULL) st->entity_name = new string(*(entity->mName));
  mInstructions.push(st);
}


void cController::newInstruction(int opcode, char* entityname)
{
  if (debug) printf("overideInstruction(%s, %s)\n", getInstructionName(opcode), entityname);

  if (entityname == NULL) return;

  rInstruction* st;

  st = new rInstruction;
  st->opcode = opcode;
  st->entity = NULL;
  st->entity_name = new string(entityname);
  mInstructions.push(st);
}


void cController::oldInstruction()
{
  if (debug) printf("oldInstruction()\n");

  // Is there any opcode to remove?
  if (mInstructions.size() == 0) return;

  rInstruction* st;

  // Remove opcode.
  st = mInstructions.top();
  if (st->entity_name != NULL) delete st->entity_name;
  mInstructions.pop();
  delete st;

  if (mInstructions.size() > 0) {
    st = mInstructions.top();
    st->entity = NULL; // Pointer might be invalid, search again.
  };
}


void cController::processSelf()
{
  if (!mEnabled) return;
  if (mDevice == NULL) return;

  if (mInstructions.empty()) {
    waitEnemy();
    return;
  };
  
  //if (mInstructions.empty()) return;
  rInstruction* st = mInstructions.top();

  if (debug && mDevice->mName != NULL) printf("Processing: %s\n", mDevice->mName->c_str());
  if (debug && st->entity_name != NULL) printf("processSelf(#%i) = <%s>\n", mInstructions.size(), st->entity_name->c_str());

  if ((st->entity == NULL) && st->entity_name != NULL)
  {
    // Search world for the entity name.
    cWorld* world = cWorld::instance;
    list<cObject*>* roles = world->filterByRole(NULL, (1UL<<cObject::MECH), false, -1);
    if (debug) world->printNames();
    if (debug) printf("|filter by role| = %i\n", roles->size());
    list<cObject*>* result = world->filterByName(mDevice, (char*) st->entity_name->c_str(), 1, roles);
    if (debug) printf("|filter by name| = %i\n", result->size());
    if (!result->empty()) {
        st->entity = (cObject*) result->front();
    } else {
        printf("no result of findName for <%s>\n", st->entity_name->c_str());
        if (debug) world->printNames(roles);
        delete result;
        delete roles;
        return;
    }
    delete result;
    delete roles;
  };

  // Set Aim Target.
  do_aimFor(st->entity);

  switch(st->opcode) {
    case WAIT: waitEnemy();      break;
    case ENGAGE: engageEnemy();    break;
    case TARGET: targetEnemy();    break;
    case ATTACK:  attackEnemy();    break;
    case FOLLOW: followLeader();   break;
    case FLEE: fleeEnemy();      break;
    case ROUTE: routeNav();       break;
    case PATROL: patrolArea();     break;
    default: printf("Invalid Instruction Request!\n");     break;
  };
}


void cController::waitEnemy()
{
  // Scan world for any enemy in range.
  cObject* nearby = enemyNearby();
  if (nearby) {
      if (mDevice->mName != NULL && nearby->mName != NULL) {
          printf("> %s Targets %s\n", mDevice->mName->c_str(), nearby->mName->c_str());
      }
      newInstruction(TARGET, nearby);
  }

  {
    do_idle();
  }
}


void cController::engageEnemy()
{
  if (inTargetRange() > 0.9)
  {
    if (mDevice->mName != NULL && mInstructions.top()->entity->mName != NULL) {
        printf("> %s Targets %s\n", mDevice->mName->c_str(), mInstructions.top()->entity->mName->c_str());
    }
    newInstruction(TARGET, mInstructions.top()->entity);
    return;
  };

  {
      do_moveTowards();
      do_aimAt();
  }
}


void cController::targetEnemy()
{
  if (inMeeleRange() > 0.9)
  {
    if (mDevice->mName != NULL && mInstructions.top()->entity->mName != NULL) {
        printf("> %s Attacks %s\n", mDevice->mName->c_str(), mInstructions.top()->entity->mName->c_str());
    }
    newInstruction(ATTACK, mInstructions.top()->entity);
    return;
  };

  if (inTargetRange() < 0.1)
  {
    //printf("inTargetRange() = %.2f\n", inTargetRange());
    if (mDevice->mName != NULL && mInstructions.top()->entity->mName != NULL) {
        printf("> %s  can't Target %s\n", mDevice->mName->c_str(), mInstructions.top()->entity->mName->c_str());
    }
    oldInstruction(); // Back to whatever was before (possibly engage or wait).
    return;
  };

  {
      do_moveTowards();
      do_aimAt();
      do_fireAt();
  }
}


void cController::attackEnemy()
{
  if (inMeeleRange() < 0.1)
  {
    if (mDevice->mName != NULL && mInstructions.top()->entity->mName != NULL) {
        printf("> %s can't Attack %s\n", mDevice->mName->c_str(), mInstructions.top()->entity->mName->c_str());
    }
    oldInstruction(); // back to target
    return;
  };

  {
      do_moveNear();
      do_aimAt();
      do_fireAt();
      //((cMech*)mDevice)->Pattern(tf, "nrnlln");
  }
}


void cController::followLeader()
{
}


void cController::fleeEnemy()
{
}


void cController::routeNav()
{
}


void cController::patrolArea()
{
}


cObject* cController::enemyNearby()
{
  return mDevice->enemyNearby();
}


float cController::inMeeleRange()
{
  return mDevice->inMeeleRange();
}


float cController::inWeaponRange()
{
  return mDevice->inWeaponRange();
}


float cController::inTargetRange()
{
  return mDevice->inTargetRange();
}


void cController::do_moveTowards()
{
  mDevice->do_moveTowards();
}


void cController::do_moveNear()
{
  mDevice->do_moveNear();
}


void cController::do_aimAt()
{
  mDevice->do_aimAt();
}


void cController::do_fireAt()
{
  mDevice->do_fireAt();
}


void cController::do_idle()
{
  mDevice->do_idle();
}


void cController::do_aimFor(cObject* target)
{
  mDevice->do_aimFor(target);
}

