// cMech.cpp

#include "cMech.h"


cMech::cMech(float* pos, float* rot) {
    mRoles |= (1UL << cObject::COLLIDEABLE);
    mRoles |= (1UL << cObject::DAMAGEABLE);
    mRoles |= (1UL << cObject::SOCIALISED);
    mRoles |= (1UL << cObject::ENTITY);
    mRoles |= (1UL << cObject::MECH);
    
    mController = new cController(this, true);
    
    if (pos != NULL) memcpy(mPos, pos, sizeof(float) * 3);
    if (rot != NULL) memcpy(mRot, rot, sizeof(float) * 3);
    memset(mTwr, 0, sizeof(float) * 3);
    
    mRotators = new float[MAX_MECH_PARTS][3];
    for (int i = 0; i < MAX_MECH_PARTS; i++) {
        memset(mRotators, 0, sizeof(float) * 3);
    };
    
    mCurrentWeapon = 0;
    mTimer = 0;
    mJetOn = 0;
    mSpeed = 0;
    
    mCameraState = 1;
    
    mModel = NULL; // Deprecated: See header.
}


void cMech::onSpawn() {
    // Find Model-Object.
    mModel = new cConstructModel((char*)"data/objects/mech0.txt");
    
    // Register position and rotation vector for Model-Object.
    mModel->setRotator(this->mRot);
    mModel->setTranslator(this->mPos);
    
    // Register some more rotation vectors for each limb you want to use.
    mModel->setConstructRotator((char*)"CTorsor", mRotators[CTorsor]);
    mModel->setConstructRotator((char*)"LUpLeg", mRotators[LUpLeg]);
    mModel->setConstructRotator((char*)"LLoLeg", mRotators[LLoLeg]);
    mModel->setConstructRotator((char*)"RUpLeg", mRotators[RUpLeg]);
    mModel->setConstructRotator((char*)"RLoLeg", mRotators[RLoLeg]);
    
    this->mountWeapon((char*)"CTorsor", &mExplosion);
    this->mWeapons.remove(&mExplosion);
    
    printf("Mech spawned\n");
    if (mModel == NULL) printf("-> warning no model\n");
}


void cMech::multEyeMatrix() {
    if (abs(mCameraState) == 1) {
        glRotatef(180, 0, 1, 0);
        if (mModel) glMultMatrixf(mModel->getConstructInvMatrix((char*)"Head"));
    } else if (abs(mCameraState) == 2) {
        glTranslatef(0, 0, -5);
        glRotatef(15, 1, 0, 0);
        glRotatef(180, 0, 1, 0);
        if (mModel) glMultMatrixf(mModel->getConstructInvMatrix((char*)"Head"));
    } else if (abs(mCameraState) == 3) {
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        glRotatef(180, 0, 1, 0);
        if (mModel) glMultMatrixf(mModel->getConstructInvMatrix((char*)"Head"));
    } else if (abs(mCameraState) == 4) {
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        glRotatef(0, 0, 1, 0);
        if (mModel) glMultMatrixf(mModel->getConstructInvMatrix((char*)"Head"));
    } else if (abs(mCameraState) == 5) {
        glTranslatef(0, 0, -100);
        glRotatef(90, 1, 0, 0);
        glRotatef(180, 0, 1, 0);
    }
    if (mModel) glMultMatrixf(mModel->getInvMatrix());
}


void cMech::Pattern(float t, char *pattern) {
    int l = strlen(pattern);
    int s = int(t/90.0) % l;
    char c = pattern[s];
    switch (c) {
        case 'l': ChassisLR(+5); break;
        case 'r': ChassisLR(-5); break;
        case 'n': break;
        default:  break;
    };
}


void cMech::ChassisLR(float r) {
    const float limit = 1.75;
    float e = r;
    if (e>+limit) e = +limit;
    if (e<-limit) e = -limit;
    mRot[Y] += e;
    mRot[Y] = fmod(mRot[Y], 360);
    if (mRot[Y] < 0) mRot[Y] += 360;
}


void cMech::ChassisUD(float r) {
    float e = r;
    if (e>+1) e = +1;
    if (e<-1) e = -1;
    mRot[X] += e;
    mRot[X] = fmod(mRot[X], 360);
    if (mRot[X] < 0) mRot[X] += 360;
}


void cMech::TowerLR(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[Y] += e;
    if (mTwr[Y] > +75) mTwr[Y] = +75;
    if (mTwr[Y] < -75) mTwr[Y] = -75;
}


void cMech::TowerUD(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[X] += e;
    if (mTwr[X] > +75) mTwr[X] = +75;
    if (mTwr[X] < -75) mTwr[X] = -75;
}


void cMech::fireAllWeapons(cObject *target) {
    list<cWeapon *>::iterator it;
    cWeapon *weapon;
    for (it = mWeapons.begin(); it != mWeapons.end(); it++) {
        weapon = *it;
        weapon->fire(target);
    };
}


void cMech::fireWeapon(int n, cObject *target) {
    int i;
    list<cWeapon *>::iterator it;
    for (i = 0, it = mWeapons.begin(); it != mWeapons.end(); it++, i++) {
        if (i == n) {
            cWeapon* weapon = *it;
            weapon->fire(target);
            break;
        }
    }
}


void cMech::mountWeapon(char* point, cWeapon *weapon) {
    if (mModel == NULL) {
        printf("mountWeapon(): no model\n");
        return;
    }
    weapon->mBasefv = mModel->getMatrix();
    weapon->mMountfv = mModel->getConstructMatrix(point);
    assert(weapon->mMountfv != NULL);
    weapon->mOwner = this;
    mWeapons.push_back(weapon);
}


void cMech::animateWalk(float spf) {
    float a = mSpeed / 100.0;
    float sp = a * spf * 5.0;
    float jt = mJetOn / 100.0 * spf * 15.0;
    const float e = 0.0174545;
    mTimer += spf;
    a = a*0.75;
    
    float vec[3] = { 0, 0, 0 };
    
    memcpy(mRotators[CTorsor], mTwr, sizeof(float) * 3);
    
    if (mPos[Y] < 0.5) {
        float t = mTimer*180;
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = (+sin(t*e)*30)*a;
        memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = (-sin(t*e)*30)*a;
        memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = (+sin(t*e*xsgn(a))*15+15)*fabs(a);
        memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = (-sin(t*e*xsgn(a))*15+15)*fabs(a);
        memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
    } else {
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = -10;
        memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = -10;
        memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = +20;
        memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);
        
        memset(vec, 0, sizeof(float) * 3);
        vec[X] = +20;
        memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
    };
    
    mPos[X] += sp*sin(mRot[Y]*e)*cos(mRot[X]*e);
    mPos[Y] += jt;
    mPos[Z] += sp*cos(mRot[Y]*e)*cos(mRot[X]*e);
    
    mPos[X] -= cWorld::instance->mGravity[X] * spf;
    mPos[Y] -= cWorld::instance->mGravity[Y] * spf;
    mPos[Z] -= cWorld::instance->mGravity[Z] * spf;
    
    // if ground contact.
    float h = checkRelHeight();
    if (h <= 0) {
        mPos[Y] -= h;
        mPos[X] += 0.05*sp*cos(mRot[Y]*e)*cos(180*mTimer*e);
        mPos[Z] += 0.05*sp*sin(mRot[Y]*e)*cos(180*mTimer*e);
    }
}


void cMech::animateSelf(float spf) {
    if (this->mHp <= 0 && mExplosion.ready()) {
        this->ChassisLR(0);
        this->TowerLR(0);
        this->TowerUD(0);
        cWorld::instance->fragObject(this);
        return;
    }
    
    if (mController && mController->mEnabled) {
        if (mPad) mPad->reset();
        mController->processSelf();
    }
    
    if (mHp <= 0) {
        mSpeed = 0;
    }
    
    if (mHp > 0 && mPad) {
        //mPad->print();
        
        this->TowerLR(50*spf * -mPad->getAxis(cPad::MECH_TURRET_LR_AXIS));
        this->TowerUD(50*spf * -mPad->getAxis(cPad::MECH_TURRET_UD_AXIS));
        
        this->ChassisLR(50*spf * -mPad->getAxis(cPad::MECH_CHASSIS_LR_AXIS));
        float tdelta = 50*spf * -mPad->getAxis(cPad::MECH_THROTTLE_AXIS);
        if (tdelta != 0.0f) {
            mSpeed += 10*tdelta;
            if (mSpeed > +220) mSpeed = +220; else
                if (mSpeed < -100) mSpeed = -100;
        }
        
        if (mPad->getButton(cPad::MECH_FIRE_BUTTON)) {
            if (true) {
                mCurrentWeapon %= mWeapons.size();
                fireWeapon(mCurrentWeapon, mTarget);
                mCurrentWeapon++;
                mCurrentWeapon %= mWeapons.size();
            } else {
                fireAllWeapons(mTarget);
            }
        }
        
        if (mPad->getButton(cPad::MECH_JET_BUTTON)) {
            mJetOn += (100-mJetOn) * 0.1;
        } else {
            mJetOn -= (mJetOn * 0.05);
        }
        
        if (mPad->getButton(cPad::MECH_CAMERA_BUTTON)) {
            // Only if Camera State is not transitional
            // then switch to next perspective on button press.
            if (mCameraState > 0) {
                mCameraState = (1 + (mCameraState % 5));
                mCameraState *= -1; // Set transitional.
            }
        } else {
            // Set Camera State as fixed.
            mCameraState = abs(mCameraState);
        }
    }
    
    animateWalk(spf);
    
    for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++) {
        cWeapon* weapon = *it;
        weapon->animate(spf);
    }
    mExplosion.animate(spf);
}


void cMech::transformSelf() {
    if (mModel) mModel->transformSelf();
    for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++) {
        cWeapon* weapon = *it;
        weapon->transform();
    };
}


void cMech::drawSolid() {
    if (mModel) mModel->drawSelf();
    for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++) {
        cWeapon* weapon = *it;
        weapon->drawSolid();
    };
    mExplosion.drawSolid();
}


void cMech::drawEffect() {
    for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++) {
        cWeapon* weapon = *it;
        weapon->drawEffect();
    };
    mExplosion.drawEffect();
}


void cMech::drawHUD() {
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);
    glLineWidth(2);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_FOG);
    glDisable(GL_DEPTH_TEST);
    
    glPushUnitOrthoProjection();
    
    glPushMatrix();
    glLoadIdentity();
    
    // Compass
    if (true) {
        glPushMatrix();
        glTranslatef(0.5, 0.97, 0);
        glScalef(1.0, 0.02, 1.0);
        glColor4f(0, 1, 0, 0.2);
        float r = this->mRot[Y];
        if (r > 180) r -= 360;
        const float d = 1.0f/360.0f;
        
        glLineWidth(3);
        glBegin(GL_LINES); {
            glColor4f(0, 0.3, 1, 0.2);
            float s = 1.2;
            for (int i = -180; i < 180; i += 10) {
                float t = fmod(r, 10);
                glVertex3f((i+t)*d, -s, 0);
                glVertex3f((i+t)*d, +s, 0);
            }
        } glEnd();
        
        glLineWidth(1);
        glBegin(GL_LINES); {
            glColor4f(1, 1, 1, 0.6);
            float s = 1.0;
            for (int i = -180; i < 180; i += 10) {
                float t = fmod(r, 10);
                glVertex3f((i+t)*d, -s, 0);
                glVertex3f((i+t)*d, +s, 0);
            }
        } glEnd();
        
        glColor4f(1, 0, 0, 0.2);
        glLineWidth(3);
        
        float w = 0.01f;
        float h = 1.0f;
        
        // N for North
        if (0) {
            glPushMatrix();
            glTranslatef(r*d,0,0);
            glBegin(GL_LINES); {
                glVertex3f(-w, +h, 0);
                glVertex3f(-w, -h, 0);
                glVertex3f(-w, +h, 0);
                glVertex3f(+w, -h, 0);
                glVertex3f(+w, +h, 0);
                glVertex3f(+w, -h, 0);
            } glEnd();
            glPopMatrix();
        }
        
        // W for West
        if (0) {
            glPushMatrix();
            glTranslatef((r-90)*d,0,0);
            glBegin(GL_LINES); {
                glVertex3f(-w, +h, 0);
                glVertex3f(-w, -h, 0);
                glVertex3f(-w, -h, 0);
                glVertex3f(+0, +h*0.5, 0);
                glVertex3f(+0, +h*0.5, 0);
                glVertex3f(+w, -h, 0);
                glVertex3f(+w, -h, 0);
                glVertex3f(+w, +h, 0);
            } glEnd();
            glPopMatrix();
        }
        
        // E for East
        if (0) {
            glPushMatrix();
            glTranslatef((r+90)*d,0,0);
            glBegin(GL_LINES); {
                glVertex3f(+w, +h, 0);
                glVertex3f(-w, +h, 0);
                glVertex3f(+w, +0, 0);
                glVertex3f(-w, +0, 0);
                glVertex3f(+w, -h, 0);
                glVertex3f(-w, -h, 0);
                glVertex3f(-w, +h, 0);
                glVertex3f(-w, -h, 0);
            } glEnd();
            glPopMatrix();
        }
                
        // S for South
        if (0) {
            glPushMatrix();
            glTranslatef((r-180)*d,0,0);
            glBegin(GL_LINES); {
                glVertex3f(+w, +h, 0);
                glVertex3f(-w, +h*0.4, 0);
                glVertex3f(-w, +h*0.4, 0);
                glVertex3f(+w, -h*0.4, 0);
                glVertex3f(+w, -h*0.4, 0);
                glVertex3f(-w, -h, 0);
            } glEnd();

            glTranslatef((+360)*d,0,0);
            glBegin(GL_LINES); {
                glVertex3f(+w, +h, 0);
                glVertex3f(-w, +h*0.4, 0);
                glVertex3f(-w, +h*0.4, 0);
                glVertex3f(+w, -h*0.4, 0);
                glVertex3f(+w, -h*0.4, 0);
                glVertex3f(-w, -h, 0);
            } glEnd();

            glPopMatrix();
        }
        
        glPopMatrix();
    }
    
    // Border
    if (true) {
        glLineWidth(5);
        glBegin(GL_LINE_STRIP); {
            glColor4f(1, 1, 0, 0.2);
            glVertex3f(0.2, 0.1, 0);
            glVertex3f(0.8, 0.1, 0);
            glVertex3f(0.9, 0.2, 0);
            glVertex3f(0.9, 0.8, 0);
            glVertex3f(0.8, 0.9, 0);
            glVertex3f(0.2, 0.9, 0);
            glVertex3f(0.1, 0.8, 0);
            glVertex3f(0.1, 0.2, 0);
            glVertex3f(0.2, 0.1, 0);
        } glEnd();
        glLineWidth(1);
        glBegin(GL_LINE_STRIP); {
            glColor4f(1, 1, 1, 0.6);
            glVertex3f(0.2, 0.1, 0);
            glVertex3f(0.8, 0.1, 0);
            glVertex3f(0.9, 0.2, 0);
            glVertex3f(0.9, 0.8, 0);
            glVertex3f(0.8, 0.9, 0);
            glVertex3f(0.2, 0.9, 0);
            glVertex3f(0.1, 0.8, 0);
            glVertex3f(0.1, 0.2, 0);
            glVertex3f(0.2, 0.1, 0);
        } glEnd();
    }
    
    float f = 250;
    
    // Tower left/right indication bar.
    glBegin(GL_LINES); {
        glColor4f(0, 0.2, 0.7, 0.4);
        glVertex3f(0.5, 0.1, 0);
        glVertex3f(0.5 + this->mTwr[1]/f, 0.1, 0);
        
        glVertex3f(0.5 + this->mTwr[1]/f, 0.08, 0);
        glVertex3f(0.5 + this->mTwr[1]/f, 0.12, 0);
        
        glVertex3f(0.5, 0.9, 0);
        glVertex3f(0.5 + this->mTwr[1]/f, 0.9, 0);
        
        glVertex3f(0.5 + this->mTwr[1]/f, 0.88, 0);
        glVertex3f(0.5 + this->mTwr[1]/f, 0.92, 0);
    } glEnd();
    
    // Tower up/down indication bar.
    glBegin(GL_LINES); {
        glColor4f(0, 0.2, 0.7, 0.4);
        glVertex3f(0.1, 0.5, 0);
        glVertex3f(0.1, 0.5 + this->mTwr[0]/f, 0);
        
        glVertex3f(0.08, 0.5 + this->mTwr[0]/f, 0);
        glVertex3f(0.12, 0.5 + this->mTwr[0]/f, 0);
        
        glVertex3f(0.9, 0.5, 0);
        glVertex3f(0.9, 0.5 + this->mTwr[0]/f, 0);
        
        glVertex3f(0.88, 0.5 + this->mTwr[0]/f, 0);
        glVertex3f(0.92, 0.5 + this->mTwr[0]/f, 0);
    } glEnd();
    
    glPopMatrix();
    
    // Weapon Status "Buttons"
    glPushMatrix();
    glLoadIdentity();
    glScalef(0.2, 0.05, 1.0);
    list<cWeapon*>::iterator i;
    for (i = this->mWeapons.begin(); i != this->mWeapons.end(); i++) {
        glLineWidth(5);
        glBegin(GL_LINE_STRIP); {
            if ((*i)->ready()) glColor4f(.4, 1, .4, 0.2);
            else glColor4f(.8, 0, 0, 0.2);
            glVertex3f(0.1, 0.1, 0);
            glVertex3f(0.9, 0.1, 0);
            glVertex3f(0.9, 0.9, 0);
            glVertex3f(0.1, 0.9, 0);
            glVertex3f(0.1, 0.1, 0);
        } glEnd();
        glLineWidth(1);
        glBegin(GL_LINE_STRIP); {
            glColor4f(1, 1, 1, 0.9);
            glVertex3f(0.1, 0.1, 0);
            glVertex3f(0.9, 0.1, 0);
            glVertex3f(0.9, 0.9, 0);
            glVertex3f(0.1, 0.9, 0);
            glVertex3f(0.1, 0.1, 0);
        } glEnd();
        glPushMatrix();
        (*i)->drawHUD();
        glPopMatrix();
        glTranslatef(0, 1, 0);
    }
    glPopMatrix();
    
    glPopProjection();
    glPopAttrib();
}


cObject::rImpact* cMech::damageByCylinder(float* a, float* b, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = mPos;
    float* r = mRot;
    glPushMatrix();
    glLoadIdentity();
    glRotatef(-r[1], 0, 1, 0);
    glTranslatef(-p[0], -p[1], -p[2]);
    float M[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, M);
    float a_local[3];
    applyMatrix3(a_local, M, a);
    float b_local[3];
    applyMatrix3(b_local, M, b);
    float mins[3] = { -1.0, 0, -0.6 };
    float maxs[3] = { +1.0, +3, +0.6 };
    float impact[3];
    float normal[3];
    float delta = collideLineWithAAB(a_local, b_local, mins, maxs, impact, normal);
    if (delta >= 0.0f) {
        printf("Mech hit\n");
        this->mHp -= damage;
        if (this->mHp <= 0) mExplosion.fire(NULL);
        result = new cObject::rImpact;
        memcpy(result->pos, impact, sizeof(float)*3);
        memset(result->nrm, 0, sizeof(float)*3);
    }
    glPopMatrix();
    return result;
}


float cMech::checkRelHeight() {
    return mPos[Y];
}


cObject* cMech::enemyNearby() {
    cObject* result = NULL;
    // Find all objects belonging to any enemy party/role.
    list<cObject*>* roles = cWorld::instance->filterByRole(this, this->mEnemies, false, -1);
    // Filter enemies by distance, choose any single one in diameter.
    list<cObject*>* scan = cWorld::instance->filterByRange(this, this->mPos, 0.0f, 50.0f, 1, roles);
    if (!scan->empty()) result = scan->front();
    delete scan;
    delete roles;
    return result;
}


float cMech::inMeeleRange() {
    if (mTarget == NULL) return 0.0f;
    float a = 16;
    float b = 24;
    float d = lineDistance(mPos, mTarget->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}


float cMech::inWeaponRange() {
    // TODO should depend on ready to fire weapons properties.
    if (mTarget == NULL) return 0.0f;
    float a = 36+10;
    float b = 44+10;
    float d = lineDistance(mPos, mTarget->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}


float cMech::inTargetRange() {
    if (mTarget == NULL) return 0.0f;
    float a = 56;
    float b = 64;
    float d = lineDistance(mPos, mTarget->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}


void cMech::do_moveTowards() {
    if (mPad == NULL) return;
    
    float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
    float  rd = 0.0f;
    rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - this->mRot[Y];
    if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, -rd);
    
    float thr = 20.0f * (1.0f - fabs(rd) / 360.0f);
    //printf("Throttle: %.2f\n", thr);
    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, (char) -thr);
}


void cMech::do_moveNear() {
    if (mPad == NULL) return;
    
    float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
    float  rd = 0.0f;
    rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - this->mRot[Y];
    if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, -rd);
    
    float d = lineDistance(this->mPos, mTarget->mPos);
    float f = (d - 11);
    if (f < -5) f = -5;
    if (f > +10) f = +10;
    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, -f*0.2f);
}


void cMech::do_aimAt() {
    if (mPad == NULL) return;
    
    float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
    float rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - (this->mRot[Y] + this->mTwr[Y]);
    if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
    mPad->setAxis(cPad::MECH_TURRET_LR_AXIS, -rd/2.0f);
}


void cMech::do_fireAt() {
    if (mPad == NULL) return;
    if (mTarget == NULL) return;
    
    float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
    float  rd = (180 + atan2(dir[X], dir[Z]) * 180.0 / M_PI) - this->mRot[Y];
    if (rd > 180) rd = 360 - rd;
    //printf("Target Angle: %.2f   Own Angle: %.2f    Angle Diff: %.2f\n", (180 + atan2(dir[X], dir[Z]) * 180.0 / M_PI), this->mRot[Y], rd);
    
    if (rand()%6 == 0) {
        if (fabs(rd) < 3) {
            mPad->setButton(cPad::MECH_FIRE_BUTTON, true);
            //printf("fire\n");
        };
    };
}


void cMech::do_idle() {
    if (mPad == NULL) return;
    mPad->reset();
}


void cMech::do_aimFor(cObject* target) {
    mTarget = target;
}

