/*
 * File:   main.cpp
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on March 28, 2008, 21:25 PM (1999)
 */

#include "main.h"



// Keyboard inputs send to which Gamepad?
cPad* gKeyPad = NULL;

// Seeing the world through which objects eyes?
cObject* gCameraObject = NULL;

// The single instance of the world we are on.
cWorld *planet = NULL;

bool paused          = false;
bool fullscreen      = false;
bool wireframe       = false;
bool nightvision     = false;


void initMission() {
    planet = new cWorld();
    planet->mMission = new cTestMission();
    gCameraObject = planet->mMission->init(planet);
    gKeyPad = gCameraObject->mPad;
}


void initGL(int width, int height) {
    glViewport(0, 0, width, height);
    
    float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
    float fogDensity = 0.001;
    
    glEnable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
    
    glClearAccum(0.0, 0.0, 0.0, 0.0);
    glClearColor(0.0, 0.125, 0.0, 0.0);
    glClearDepth(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT);
    
    glShadeModel(GL_SMOOTH);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);
    
    glFogi(GL_FOG_MODE, GL_EXP);
    glFogfv(GL_FOG_COLOR, fogColor);
    glFogf(GL_FOG_DENSITY, fogDensity);
    
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    // Lighting
    glEnable(GL_LIGHTING);
    float p[] = { 0, 1000, 0, 1 };
    float a[] = { 0.5, 0.5, 0.5, 1 };
    glLightfv(GL_LIGHT0, GL_POSITION, p);
    glLightfv(GL_LIGHT0, GL_AMBIENT, a);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.001);
    glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.00001);
    glEnable(GL_COLOR_MATERIAL);
}


void drawFrame(int elapsed_msec) {
    //printf("elapsed: %.5f\n", elapsed_msec / 1000.0f);
    //if (hurlmotion) elapsed = int(elapsed * 1.0f);
    
    // calculate the time that the last frame took.
    planet->advanceTime(elapsed_msec);
    
    // Delete Fragged Objects of previous frame.
    planet->bagFragged();
    
    // Calculate new animation state and the like (if not paused).
    if (!paused)
        planet->animateObjects();
    
    // Calculate positions through transformations.
    glPushMatrix();
    glLoadIdentity();
    planet->transformObjects();
    glPopMatrix();
    
    // Draw fewer frames?
    //static int cnt = 0;
    //cnt++;
    //if (cnt % 2 == 0) return;
    
    // SOLID / WIREFRAME drawing scheme.
    if (!wireframe) {
        float dark[] = { 0.0, 0.0, 0.0, 1.0 };
        glClearColor(0.2, 0.2, 0.2, 1.0);
        glEnable(GL_FOG);
        glFogfv(GL_FOG_COLOR, dark);
        //glClearColor(dome->mLoColor[X], dome->mLoColor[Y], dome->mLoColor[Z], 1.0);
        //glEnable(GL_FOG);
        //glFogfv(GL_FOG_COLOR, dome->mLoColor);
        glEnable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_CULL_FACE);
        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
    } else {
        float dark[] = { 0.0, 0.0, 0.0, 1.0 };
        glClearColor(0.2, 0.2, 0.2, 1.0);
        glEnable(GL_FOG);
        glFogfv(GL_FOG_COLOR, dark);
        glDisable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glDisable(GL_CULL_FACE);
    };
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPushPerspectiveProjection(60);
    
    // Setup camera.
    glLoadIdentity();
    if (gCameraObject) gCameraObject->multEyeMatrix();
    
    // Draw the surounding Sky- and Ground-Dome/Box
    if (planet->mBackground) planet->mBackground->drawBackground(planet->mTime);
    
    //glBindTexture(GL_TEXTURE_2D, gAttentionTextureBind);
    
    // Draw the Objects themselves.
    planet->drawSolid();
    
    // Draw the Object's translucent effects.
    planet->drawEffect();
    
    // Draw the Head-Up-Display of the currently spectating Object.
    gCameraObject->drawHUD();
    
    glPopProjection();
    
    glFlush();
    
    if (nightvision) {
        glAccumBlurInverse(0.83);
    }
}


#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include <SDL/SDL_timer.h>


void updateKey(Uint8 keysym) {
    if (keysym == _WIREFRAME_KEY) {
        wireframe = !wireframe;
    } else if (keysym == _NIGHTVISION_KEY) {
        nightvision = !nightvision;
    } else if (keysym == _PAUSE_KEY) {
        paused = !paused;
    };
}

void updatePad(cPad* pad, SDL_Joystick* joy, int* mapping) {
    if (pad == NULL) return;

    static int default_mapping[] = {
        0,  1,   2,  3,
        4,  5,   6,  7,
        8,  9,  10, 11,
        12, 13, 14, 15,
        16, 17, 18, 19,
        20, 21, 22, 23,
        24, 25, 26, 27,
        28, 29, 30, 31
    };
    int* map = default_mapping;
    if (mapping != NULL) map = mapping;
    
    pad->reset();
    if (joy != NULL) {
        loopi(4, pad->setAxis((cPad::Axes) i, SDL_JoystickGetAxis(joy, i)/32000.0f) );
        loopi(32, if (SDL_JoystickGetButton(joy, map[i])) pad->setButton((cPad::Buttons)i, true) );
        //pad->print();
    };
    
    Uint8 *keystate = SDL_GetKeyState(NULL);
    if (keystate == NULL) return;
    // First Analogue Control Stick
    if (keystate[SDLK_LEFT] == 1) pad->setAxis(cPad::AX_LR1, -1.0f);
    if (keystate[SDLK_RIGHT] == 1) pad->setAxis(cPad::AX_LR1, +1.0f);
    if (keystate[SDLK_UP] == 1) pad->setAxis(cPad::AX_UD1, -1.0f);
    if (keystate[SDLK_DOWN] == 1) pad->setAxis(cPad::AX_UD1, +1.0f);
    // Start and Select
    if (keystate[SDLK_RETURN] == 1) pad->setButton(cPad::BT_START, true);
    if (keystate[SDLK_BACKSPACE] == 1) pad->setButton(cPad::BT_SELECT, true);
    // Lx and Rx Buttons
    if (keystate[SDLK_w] == 1) pad->setButton(cPad::BT_L1, true);
    if (keystate[SDLK_q] == 1) pad->setButton(cPad::BT_L2, true);
    if (keystate[SDLK_r] == 1) pad->setButton(cPad::BT_R1, true);
    if (keystate[SDLK_t] == 1) pad->setButton(cPad::BT_R2, true);
    // Primary Buttons
    if (keystate[SDLK_s] == 1) pad->setButton(cPad::BT_SQUARE, true);
    if (keystate[SDLK_e] == 1) pad->setButton(cPad::BT_DELTA, true);
    if (keystate[SDLK_d] == 1) pad->setButton(cPad::BT_CROSS, true);
    if (keystate[SDLK_f] == 1) pad->setButton(cPad::BT_CIRCLE, true);
    // Second Analogue Stick or HUD:
    if (keystate[SDLK_LALT] == 0) {
        if (keystate[SDLK_j]) gKeyPad->setAxis(cPad::AX_LR2, -1.0f);
        if (keystate[SDLK_l]) gKeyPad->setAxis(cPad::AX_LR2, +1.0f);
        if (keystate[SDLK_i]) gKeyPad->setAxis(cPad::AX_UD2, -1.0f);
        if (keystate[SDLK_k]) gKeyPad->setAxis(cPad::AX_UD2, +1.0f);
    } else {
        if (keystate[SDLK_j]) gKeyPad->setButton(cPad::BT_HL, true);
        if (keystate[SDLK_l]) gKeyPad->setButton(cPad::BT_HR, true);
        if (keystate[SDLK_i]) gKeyPad->setButton(cPad::BT_HU, true);
        if (keystate[SDLK_k]) gKeyPad->setButton(cPad::BT_HD, true);
    };
    // Stick Buttons
    if (keystate[SDLK_PERIOD]) gKeyPad->setButton(cPad::BT_J1B, true);
    if (keystate[SDLK_COMMA]) gKeyPad->setButton(cPad::BT_J2B, true);
}

#include <iostream>
using namespace std;

void printHelp() {
    cout << endl;
    cout << "LinWarrior" << endl;
    cout << "  by Benjamin Pickhardt (pickhardt debitel.net" << endl;
    cout << endl;
    cout << "Build:" << endl;
    cout << "  " << __DATE__ << endl;
    cout << "  " << __TIME__ << endl;
    cout << endl;
    cout << "Options:" << endl;
    cout << endl;
    cout << "  --help          prints this help screen" << endl;
    cout << "  --resolution    followed by resolution WIDTH or WIDTHxHEIGHT" << endl;
    cout << "  --window        start in windowed mode" << endl;
    cout << "  --joypad        followed by a (default) or b." << endl;
    cout << "                  The a-Type Gamepad is noticable by zedwise" << endl;
    cout << "                  ordered buttons and b-Type by clockwise buttons" << endl;
    //cout << "  --mission       followed by mission name (todo)" << endl;
    //cout << "  --music         play named title overriding mission music (???)" << endl;
    cout << endl;
    cout << "  All options can be abbreviated by their first letter:" << endl;
    cout << "    --help is the same as -h" << endl;
    cout << endl;
    cout << endl;
    exit(1);
}

#define argcmp(p, b)  ( (strcmp(p,b)==0) || (b[0]=='-' && b[1]==p[2] && b[2]=='\0') )
#define arg_i(p)      argcmp(p, args[i])
#define advance_i()   { i++; if (i > argc) { printHelp; } }

int sdlmain(int argc, char** args) {
    bool printpad = false;
    float fps = 40.0f;
    bool fullscreen = true;
    int width = 800;
    int height = 600;
    int* joymap = map_zedwise;
    
    for (int i = 0; i < argc; i++) {
        if (arg_i("--help")) {
            printHelp();
        } else if (arg_i("--resolution")) {
            advance_i();
            string s = args[i];
            int x0 = s.find('x',  0);
            int x1 = s.find('x', x0+1);
            // Is it WIDTHxHEIGHT? or else just Width?
            if (x0 > 0) {
                // Width
                string w = s.substr(0, x0-0);
                width = atoi(w.c_str());
                // Height
                string h = s.substr(x0+1, x1-(x0+1));
                height = atoi(h.c_str());
                // Depth
                if (x1 > 0) {
                    string d = s.substr(x1+1);
                    int depth = atoi(d.c_str());
                    // unused value
                }
            } else {
                // Just take the width and find a fitting height.
                width = atoi(s.c_str());
                double aspect = 4.0f / 3.0f;
                if (width > 1024) aspect = 16.0f / 9.0f;
                height = (int) (width / aspect);
                if (height % 2 == 1) height += 1;
            }
            cout << "Resolution: " << width << " x " << height << endl;
        } else if (arg_i("--joypad")) {
            advance_i();
            string s = args[i];
            if (s.compare("a")) {
                joymap = map_zedwise;
            } else if (s.compare("b")) {
                joymap = map_clockwise;
            } else {
                printHelp();
            }
        } else if (arg_i("--window")) {
            fullscreen = false;
        }
    }
    
    float spf = 1.0f / fps;
    
    // Iitialize SDL video, audio and input
    //if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK ) < 0 ) {
    if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ) < 0 ) {
        printf( "Unable to init SDL: %s\n", SDL_GetError() );
        return 1;
    }
    
    // Cleanup afterwards.
    atexit(SDL_Quit);

    /*
    int audio_rate = (1<<1)*11025;
    Uint16 audio_format = AUDIO_S16SYS;
    int audio_channels = 2; // 1 == mono, 2 == stereo
    int audio_buffers = (1<<0)*1024;
    
    if(Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers) != 0) {
        fprintf(stderr, "Unable to initialize audio: %s\n", Mix_GetError());
        return 1;
    }
    */
    
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1);
    SDL_Surface* screen = NULL;
    if (fullscreen) screen = SDL_SetVideoMode(width, height, 0, SDL_OPENGLBLIT | SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
    else screen = SDL_SetVideoMode(width, height, 0, SDL_OPENGLBLIT | SDL_HWSURFACE | SDL_DOUBLEBUF);
    if ( !screen ) {
        printf("Unable to set %i x %i video: %s\n", width, height, SDL_GetError());
        return 1;
    }
    
    if (false) {
        string glinfo;
        glinfo = (const char*)glGetString(GL_RENDERER);
        cout << glinfo << endl;
        glinfo = (const char*)glGetString(GL_VENDOR);
        cout << glinfo << endl;
        glinfo = (const char*)glGetString(GL_VERSION);
        cout << glinfo << endl;
        glinfo = (const char*)glGetString(GL_EXTENSIONS);
        cout << glinfo << endl;
    }
    
    
    initGL(screen->w, screen->h);
    initMission();
    
    SDL_ShowCursor(SDL_DISABLE);
    SDL_Joystick* joy0 = SDL_JoystickOpen(0);
    
    // program main loop
    bool done = false;
    while (!done) {
        int start = SDL_GetTicks();
        
        // message processing loop
        SDL_Event event;
        while (SDL_PollEvent(&event)) {
            // check for messages
            switch (event.type) {
                // exit if the window is closed
                case SDL_QUIT: {
                    done = true;
                    break;
                }
                // check for keypresses
                case SDL_KEYDOWN: {
                    // exit if ESCAPE is pressed
                    updateKey(event.key.keysym.sym);
                    if (event.key.keysym.sym == SDLK_ESCAPE) done = true;
                    break;
                }
            } // switch
        } // while poll event
        
        // Possibly printout info on pushed buttons.
        if (printpad) loopi(32, if (SDL_JoystickGetButton(joy0, i)) printf("Button %i\n", i) );
        // Transfer real Joystick-/Gamepad-Input into the virtual gamepad
        // of the player by using a (re-)mapping.
        updatePad(gKeyPad, joy0, joymap);
        // After securing Matrix update world and draw frame.
        glPushMatrix();
            drawFrame(spf * 1000.0f);
        glPopMatrix();
        
        //SDL_Rect r = { 10,10, 50,50 };
        //SDL_FillRect(screen, &r, SDL_MapRGB(screen->format, 0, 0, 0));
        //SDL_FillRect(screen, 0, SDL_MapRGB(screen->format, 0, 0, 0));
        //SDL_BlitSurface(bmp, 0, screen, &dstrect);
        //SDL_Flip(screen);
        
        SDL_GL_SwapBuffers();
        
        // Lock framerate and give back idle time to the os,
        // other processes or threads.
        spf = (1.0f/fps);
        int delta = SDL_GetTicks() - start; // msec since frame started
        int delay = ((int)(spf * 1000.0f)) - delta;
        if (delay > 0) SDL_Delay(delay);
    } // end main loop
    
    printf("Thank you for playing.\n");
    return 0;
}



//
// GOD begins below!
//

#ifdef _WIN32
extern "C" int __stdcall WinMain(void* hInstance, void* hPrevInstance, char* lpCmdLine, int nCmdShow) {
    int    argc = 2;
    char  *argv[] = { "Nothing", "Nothing", "Nothing" };
#else
    int main(int argc, char **args) {
#endif
        
#if 0
        return glutmain(argc, args);
#else
        return sdlmain(argc, args);
#endif
    }
    
